/*
 * Copyright (c) 1997-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2005      Arun M. Krishnakumar,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2004      Abhishek Rai,
 * Copyright (c) 2005      Sunil Satnur,
 * Copyright (c) 2006      Avishay Traeger,
 * Copyright (c) 2001-2006 Stony Brook University,
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: main.c,v 1.52 2006/12/31 01:43:58 crpatti Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "raif.h"
#include "ECC/levels.h"
#include <linux/module.h>

/* This definition must only appear after we include <linux/module.h> */
#ifndef MODULE_LICENSE
#define MODULE_LICENSE(foo)
#endif /* not MODULE_LICENSE */

/* sb we pass is raif's super_block */
int raif_interpose_debug(dentry_t *dentry, super_block_t *sb, int flag, char *file, int line, const char *function)
{
	inode_t *hidden_inode;
	dentry_t *hidden_dentry;
	int err = 0;
	inode_t *inode;
	int bindex, bnum;

	print_entry_location();

	bnum = dbnum(dentry);

	for (bindex = 0; bindex < bnum; bindex++) {
		PASSERT2(dtohd_index(dentry, bindex));
		if (!dtohd_index(dentry, bindex)->d_inode) {
			printk("<0>dtohd_index(dentry, bindex)->d_inode NULL for [%s, 0x%x]\n",dentry->d_name.name, dtopd(dentry)->mask); 
		}
		PASSERT2(dtohd_index(dentry, bindex)->d_inode);
	}

	/* We allocate our new inode below, by calling iget.
	 * iget will call our read_inode which will initialize some
	 * of the new inode's fields */
	/*for (bindex = 0; bindex < bnum; bindex++) {
		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry) {
			continue;
		}

		// check that the lower file system didn't cross a mount point
		hidden_inode = hidden_dentry->d_inode;
		if (hidden_inode && (hidden_inode->i_sb->s_dev != stohs_index(sb, global_bindex)->s_dev)) {
			err = -EXDEV;
			goto out;
		}
	}*/

	/* On revalidate we've already got our own inode and just need to fix it up. */
	if (flag != INTERPOSE_REVAL) {
		/* get unique inode number for raif */
		inode = iget(sb, iunique(sb, RAIF_ROOT_INO));
		if (!inode) {
			err = -EACCES;		/* should be impossible??? */
			goto out;
		}
	} else {
		inode = dentry->d_inode;
		PASSERT(inode);
		itohi_ptr(inode) = KMALLOC(sizeof(inode_t *) * sbnum(sb), GFP_KERNEL);
		if (!itohi_ptr(inode)) {
			err = -ENOMEM;
			goto out;
		}
		init_itohi_ptr(inode, sbnum(sb));
	}

	for (bindex = 0; bindex < bnum; bindex++) {
		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry)
			itohi_index(inode, bindex) = NULL;
		else if (hidden_dentry->d_inode != NULL)
			itohi_index(inode, bindex) = igrab(hidden_dentry->d_inode);
	}

	ibnum(inode) = bnum;

	/* Use attributes from the first branch. */
	hidden_inode = itohi_index(inode, 0);

	/* Use different set of inode ops for symlinks & directories*/
	if (S_ISLNK(hidden_inode->i_mode))
		inode->i_op = &raif_symlink_iops;
	else if (S_ISDIR(hidden_inode->i_mode))
		inode->i_op = &raif_dir_iops;

	/* Use different set of file ops for directories */
	if (S_ISDIR(hidden_inode->i_mode))
		inode->i_fop = &raif_dir_fops;

	/* properly initialize special inodes */
	if (S_ISBLK(hidden_inode->i_mode) || S_ISCHR(hidden_inode->i_mode) ||
			S_ISFIFO(hidden_inode->i_mode) || S_ISSOCK(hidden_inode->i_mode)) {
		init_special_inode(inode, hidden_inode->i_mode, hidden_inode->i_rdev);
	}

	/* only (our) lookup wants to do a d_add */
	switch(flag) {
		case INTERPOSE_DEFAULT:
			d_instantiate(dentry, inode);
			break;
		case INTERPOSE_LOOKUP:
			d_add(dentry, inode);
			break;
		case INTERPOSE_REVAL:
			/* Do nothing. */
			break;
		default:
			FISTBUG("Invalid interpose flag passed!");
	}

	PASSERT(dtopd(dentry));

	/* all well, copy inode attributes */
	fist_copy_attr_allbutsizes(inode, hidden_inode);
	
#ifdef RAIF_TAG
	if (inode) {
		err = raif_rule_interpose(dentry);
		if (err)
			goto out;
	}
#endif
	update_inode_size(inode);
out:
	print_exit_status(err);
	return err;
}

/*
 * Parse mount options:
 * Current mount command example = 
 * mount -t raif -o dirs=b1:b2:b3:b4,debug=N null /mnt/raif
 * any leading and/or trailing colons are ignored
 *
 * Returns the dentry object of the lower-level (hidden) directory;
 * We want to mount our stackable file system on top of that hidden directory.
 *
 * Sets default debugging level to N, if any.
 */
struct raif_dentry_info * raif_parse_options(super_block_t *sb, char *options)
{
	struct raif_dentry_info *hidden_root_info;
	struct nameidata nd;
	char *name, *tmp, *tmp2, *end, *end2;
	int err = 0;
	int bindex = 0;

	print_entry_location();

	/* allocate private data area */
	hidden_root_info = KMALLOC(sizeof(struct raif_dentry_info), GFP_KERNEL);
	if (!hidden_root_info) {
		err = -ENOMEM;
		goto out_error;
	}
	init_dentry_info(hidden_root_info);

	/* We don't want to go off the end of our arguments later on. */
	for (end = options; *end; end++);

	while (options < end) {
		tmp = options;
		while (*tmp && *tmp != ',') {
			tmp++;
		}
		*tmp = '\0';

		if (!strncmp("dirs=", options, 5)) {

			options = options + 5;

			/* We don't want to go off the end of our dir arguments later on. */
			for (end2 = options; *end2; end2++) ;

			/* colon at the beginning */
			if (*options == ':') {
				options++;
			}
			/* colon at the end */
			if (*(end2-1) == ':') {
				end2--;
				*end2 = '\0';
			} 

			/* counting the number of intermediate colons */
			for(tmp2 = options; *tmp2; tmp2++) {
				if(*tmp2 == ':') {
					sbnum(sb)++;
				}
			}
			/* increment by 1 more because actual number of branches is 1 more,  b1:b2:b3 */
			sbnum(sb)++;

			/* allocate space for  underlying pointers to hidden dentry */
			hidden_root_info->udi_dentry = KMALLOC(sizeof(dentry_t *) * sbnum(sb), GFP_KERNEL);
			if (!hidden_root_info->udi_dentry) {
				err = -ENOMEM;
				goto out_error;
			}
			memset(hidden_root_info->udi_dentry, 0, sizeof(dentry_t *) * sbnum(sb));

			/* allocate space for underlying pointers to super block */
			stohs_ptr(sb) = KMALLOC(sizeof(super_block_t *) * sbnum(sb), GFP_KERNEL);
			if (!stohs_ptr(sb)) {
				err = -ENOMEM;
				goto out_error;
			}
			init_stohs_ptr(sb, sbnum(sb));

			/* allocate space for array of pointers for underlying mount points */
			stohiddenmnt_ptr(sb) = KMALLOC(sizeof(struct vfsmount *) * sbnum(sb), GFP_KERNEL);
			if (!stohiddenmnt_ptr(sb)) {
				err = -ENOMEM;
				goto out_error;
			}
			init_stohiddenmnt_ptr(sb, sbnum(sb));

#ifdef RAIF_LB
			stopd(sb)->load = KMALLOC(sizeof(int) * sbnum(sb), GFP_KERNEL);
			stopd(sb)->scratchpool = KMALLOC(sizeof(char) * sbpoolsize(sb), GFP_KERNEL);
			stopd(sb)->pool = KMALLOC(sizeof(char) * sbpoolsize(sb), GFP_KERNEL);
			stopd(sb)->randbuf = KMALLOC(sizeof(char) * sbpoolsize(sb), GFP_KERNEL);

#ifdef LB_ALTERNATE
			// FIXME: should be sbstart(sb) <but sbstart not yet initialized
			stopd(sb)->curr = 0;
#endif /* LB_ALTERNATE */

			/* lbfs: allocate space for per-branch latency / lat-lock structs */
			stopd(sb)->lat = KMALLOC(sizeof(latency_t) * sbnum(sb), GFP_KERNEL);
			stopd(sb)->lat_lock = KMALLOC(sizeof(spinlock_t) * sbnum(sb), GFP_KERNEL);
			if (!stopd(sb)->lat || !stopd(sb)->lat_lock) {
				err = -ENOMEM;
				goto out_error;
			}
			memset(stopd(sb)->pool, 0, sizeof(char) * sbpoolsize(sb));	/* 0xff is reserved */
			memset(stopd(sb)->scratchpool, 0, sizeof(char) * sbpoolsize(sb));	/* 0xff is reserved */
			memset(stopd(sb)->load, 0, sizeof(int) * sbnum(sb));
			memset(stopd(sb)->lat, 0, sizeof(latency_t) * sbnum(sb));
			memset(stopd(sb)->lat_lock, 0, sizeof(spinlock_t) * sbnum(sb));
			init_pool(sb);
#endif /* RAIF_LB */

			/* now parsing the string b1:b2(=rw):b3(=ro):b4 */
			while (options < end2) {
				int perms;

				int l;

				tmp2 = options;

				while (*tmp2 && *tmp2 != ':') {
					tmp2++;
				}
				*tmp2 = '\0';

				/* name contains individual dir */
				name = options;

				/* strip off =rw or =ro if it is specified. */
				l = strlen(name);
				if(!strcmp(name + l - 3, "=ro")) {
					perms = MAY_READ;
					name[l - 3] = '\0';
				} else if(!strcmp(name + l - 3, "=rw")) {
					perms = MAY_READ | MAY_WRITE;
					name[l - 3] = '\0';
				} else {
					perms = MAY_READ | MAY_WRITE;
				}

				fist_dprint(4, "raif: using directory: %s (%c%c)\n", name, perms & MAY_READ ? 'r' : '-', perms & MAY_WRITE ? 'w' : '-');
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
				if (path_init(name, LOOKUP_FOLLOW, &nd))
					err = path_walk(name, &nd);
#else
				err = path_lookup(name, LOOKUP_FOLLOW, &nd);
#endif
				if (err) {
					printk("raif: error accessing hidden directory '%s'\n", name);
					goto out_error;
				}

				hidden_root_info->udi_dentry[bindex] = nd.dentry;
				stopd(sb)->usi_hidden_mnt[bindex] = nd.mnt;

				// increment end branch as we add more branches
				hidden_root_info->b_num = bindex + 1;

				// increment the branch index for the next branch to be added
				bindex++;

				/* increment the string pointer */
				options = tmp2 + 1;
			}
		} else if (!strncmp("debug=", options, 6)) {
			char *endptr;
			int debug = simple_strtoul(options + 6, &endptr, 0);
			if (*endptr) {
				printk(KERN_WARNING "raif: invalid debug option '%s'\n", options);	
			}
			fist_set_debug_value(debug);
#ifdef RAIF_TAG
#endif
		} else {
			printk(KERN_WARNING "raif: unrecognized option '%s'\n", options);
		}
		options = tmp + 1;
	}

#ifdef RAIF_LB
	kernel_thread(load_update, (void *)sb, CLONE_VM | CLONE_FS | CLONE_FILES);
	//kernel_thread(load_update, (void *)sb, 0);
	atomic_set(&stopd(sb)->die, 0);
	init_waitqueue_head(&(stopd(sb)->clean));
#endif /* RAIF_LB */

	goto out;

out_error:
	if (stohiddenmnt_ptr(sb)) {
		KFREE(stohiddenmnt_ptr(sb));
	}
	if (stohs_ptr(sb)) {
		KFREE(stohs_ptr(sb));
	}

#ifdef RAIF_LB
	if (stopd(sb)->scratchpool) {
		KFREE(stopd(sb)->scratchpool);
	}
	if (stopd(sb)->pool) {
		KFREE(stopd(sb)->pool);
	}
	if (stopd(sb)->randbuf) {
		KFREE(stopd(sb)->randbuf);
	}
	if (stopd(sb)->load) {
		KFREE(stopd(sb)->load);
	}
	if (stopd(sb)->lat) {
		KFREE(stopd(sb)->lat);
	}
	if (stopd(sb)->lat_lock) {
		KFREE(stopd(sb)->lat_lock);
	}
#endif /* RAIF_LB */

	if (hidden_root_info) {
		if (hidden_root_info->udi_dentry) {
			KFREE(hidden_root_info->udi_dentry);
		}
		KFREE(hidden_root_info);
	}
	hidden_root_info = ERR_PTR(err);
out:
	print_exit_location();
	return hidden_root_info;
}

#ifdef FIST_MALLOC_DEBUG
/* for malloc debugging */
atomic_t raif_malloc_counter;
atomic_t raif_mallocs_outstanding;

atomic_t raif_dget_counter;
atomic_t raif_dgets_outstanding;

void * raif_kmalloc(size_t len, int flag, int line, const char *file)
{
	void *ptr = (void *) kmalloc(len, flag);
	if (ptr) {
		atomic_inc(&raif_malloc_counter);
		atomic_inc(&raif_mallocs_outstanding);
		printk("KM:%d:%d:%p:%d:%s\n", atomic_read(&raif_malloc_counter),atomic_read(&raif_mallocs_outstanding), ptr, line, file);
	}
	return ptr;
}

void raif_kfree(void *ptr, int line, const char *file)
{
	atomic_inc(&raif_malloc_counter);
	atomic_dec(&raif_mallocs_outstanding);
	printk("KF:%d:%d:%p:%d:%s\n", atomic_read(&raif_malloc_counter), atomic_read(&raif_mallocs_outstanding), ptr, line, file);
	kfree(ptr);
}

#endif /* FIST_MALLOC_DEBUG */

/* for attach mode, we use a different ->read_super() in attach.c */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
super_block_t *
#else
static int
#endif
raif_read_super(super_block_t *sb, void *raw_data, int silent)
{
	super_block_t *ret_sb = NULL;
	super_block_t *temp_sb = NULL;
	struct raif_dentry_info *hidden_root_info = NULL;
	int global_bindex, bnum;
	unsigned long long maxbytes;
	int err = 0;

	print_entry_location();

#ifdef FIST_MALLOC_DEBUG
	atomic_set(&raif_malloc_counter, 0); /* for malloc debugging */
#endif /* FIST_MALLOC_DEBUG */

	if (!raw_data) {
		printk(KERN_WARNING "raif_read_super: missing data argument\n");
		err = -EINVAL;
		goto out;
	}

	/*
	 * Allocate superblock private data
	 */
	stopd_lhs(sb) = KMALLOC(sizeof(struct raif_sb_info), GFP_KERNEL);
	if (!stopd(sb)) {
		printk(KERN_WARNING "%s: out of memory\n", __FUNCTION__);
		err = -ENOMEM;
		goto out;
	}
	init_stopd(sb);

	hidden_root_info = raif_parse_options(sb, raw_data);

	if (IS_ERR(hidden_root_info)) {
		hidden_root_info = NULL;
		err = PTR_ERR(hidden_root_info);
		printk(KERN_WARNING "raif_read_super: lookup_dentry failed (err = %ld)\n", PTR_ERR(hidden_root_info));
		goto out_free;
	}

	/* 
	 * Traverse the entire returned hidden_root_info structure
	 * to find if all inodes exist
	 */
	bnum = hidden_root_info->b_num;
	for (global_bindex = 0; global_bindex < bnum; global_bindex++) {
		if (!hidden_root_info->udi_dentry[global_bindex]->d_inode) {
			printk(KERN_WARNING "raif_read_super: no directory to interpose on\n");
			err = -ENOENT;
			goto out_dput;
		}
	}

	/* 
	 * set the hidden superblock field of upper superblock 
	 */
	sbnum(sb) = bnum;
	for (global_bindex = 0; global_bindex < bnum; global_bindex++) {
		temp_sb = hidden_root_info->udi_dentry[global_bindex]->d_sb;
		stohs_index(sb, global_bindex) = temp_sb;
		if (!strncmp(temp_sb->s_type->name, "nfs", 3))
			stopd(sb)->is_lower_nfs = 1;
	}

	maxbytes = -1;
	for (global_bindex = 0; global_bindex < bnum; global_bindex++) {
		if (maxbytes < stohs_index(sb, global_bindex)->s_maxbytes) {
			maxbytes = stohs_index(sb, global_bindex)->s_maxbytes;
		}
	}
	sb->s_maxbytes = maxbytes;

	sb->s_op = &raif_sops;

/*#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	sb->s_root = d_alloc(NULL, &(const struct qstr) { "/", 1, 0 });
#else*/
	sb->s_root = d_alloc(NULL, &(const struct qstr){hash: 0, name: "/", len: 1});
//#endif
	if (IS_ERR(sb->s_root)) {
		err = PTR_ERR(sb->s_root);
		printk(KERN_WARNING "raif_read_super: d_alloc failed\n");
		goto out_dput;
	}

	sb->s_root->d_op = &raif_dops;
	sb->s_root->d_sb = sb;
	sb->s_root->d_parent = sb->s_root;

	/* link the upper and lower dentries */
	/* dtopd(sb->s_root) */
	sb->s_root->d_fsdata = KMALLOC(sizeof(struct raif_dentry_info), GFP_KERNEL);
	if (!dtopd(sb->s_root)) {
		err = -ENOMEM;
		goto out_dput2;
	}
	memset(sb->s_root->d_fsdata, '\0', sizeof(struct raif_dentry_info));

	/* allocate array of dentry pointers for the root dentry */
	dtohd_ptr(sb->s_root) = KMALLOC(sizeof(dentry_t *) * sbnum(sb), GFP_KERNEL);
	if (!dtohd_ptr(sb->s_root)) {
		err = -ENOMEM;
		goto out_dput3;
	}
	init_dtohd_ptr(sb->s_root, sbnum(sb));

	/* Set the hidden dentries for s_root */
	for (global_bindex = 0; global_bindex < bnum; global_bindex++) {
		dtohd_index(sb->s_root, global_bindex) = hidden_root_info->udi_dentry[global_bindex];
	}
	dbnum(sb->s_root) = bnum;


#ifdef RAIF_TAG
	// XXX parse top directory rules here
	dtopd(sb->s_root)->d.rules.rule = NULL;
#endif

	if ((err = raif_interpose(sb->s_root, sb, INTERPOSE_DEFAULT)) != 0) {
		// XXX something is wrong with this error handling code
		goto out_dput4;
	}

	ret_sb = sb;

	goto out;

out_dput4:
	KFREE(dtohd_ptr(sb->s_root));
out_dput3:
	KFREE(dtopd(sb->s_root));
out_dput2:
	dput(sb->s_root);
out_dput:
	for (global_bindex = 0; global_bindex < bnum; global_bindex++) {
		if (hidden_root_info->udi_dentry[global_bindex]) {
			dput(hidden_root_info->udi_dentry[global_bindex]);
		}
		if (stopd(sb) && stohm_index(sb, global_bindex)) {
			mntput(stohm_index(sb, global_bindex));
		}
	}
out_free:
	KFREE(stopd(sb));
	stopd_lhs(sb) = NULL;
out:
	if (hidden_root_info) {
		if (hidden_root_info->udi_dentry) {
			KFREE(hidden_root_info->udi_dentry);
		}
		KFREE(hidden_root_info);
	}
	print_exit_location();

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if(err)
		return(ERR_PTR(err));
	else
		return ret_sb;
#else
	return(err);
#endif
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static DECLARE_FSTYPE(raif_fs_type, "raif", raif_read_super, 0);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
static struct super_block *raif_get_sb(struct file_system_type *fs_type,
		int flags, const char *dev_name,
		void *raw_data)
{
	return get_sb_nodev( fs_type, flags, raw_data, raif_read_super);
}
#else
static int raif_get_sb(struct file_system_type *fs_type,
		int flags, const char *dev_name,
		void *raw_data, struct vfsmount *mnt)
{
	return get_sb_nodev( fs_type, flags, raw_data, raif_read_super, mnt);
}
#endif

void raif_kill_block_super(struct super_block *sb)
{
	generic_shutdown_super(sb);
}

static struct file_system_type raif_fs_type = {
	.owner          = THIS_MODULE,
	.name           = "raif",
	.get_sb         = raif_get_sb,
	.kill_sb        = raif_kill_block_super,
	.fs_flags       = 0,
};

#endif

static int __init init_raif_fs(void)
{
	int err = 0;

	/*Init all Caches*/
	if ((err = init_filldir_cache()))
		return err;
	printk("Registering raif version $Id: main.c,v 1.52 2006/12/31 01:43:58 crpatti Exp $\n");
	return register_filesystem(&raif_fs_type);
}

static void __exit exit_raif_fs(void)
{
	destroy_filldir_cache();
	printk("Unregistering raif version $Id: main.c,v 1.52 2006/12/31 01:43:58 crpatti Exp $\n");
	unregister_filesystem(&raif_fs_type);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
EXPORT_NO_SYMBOLS;
#endif
MODULE_AUTHOR("Nikolai Joukov <kolya@cs.sunysb.edu> and Erez Zadok <ezk@cs.sunysb.edu>");
MODULE_DESCRIPTION("FiST-generated raif filesystem");
MODULE_LICENSE("GPL");

/* XXX: are we using this module parameter?? commented it out
 * because there are problems with MODULE_PARM in 2.6.19 */
/*MODULE_PARM(fist_debug_var, "i");
MODULE_PARM_DESC(fist_debug_var, "Debug level");*/

module_init(init_raif_fs)
module_exit(exit_raif_fs)

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
